@@ -43,28 +43,19 @@
 
 
 /**
- * SECTION:gnetworkservice
- * @short_description: A GSocketConnectable for resolving SRV records
- * @include: gio/gio.h
+ * GNetworkService:
  *
- * Like #GNetworkAddress does with hostnames, #GNetworkService
+ * Like [class@Gio.NetworkAddress] does with hostnames, `GNetworkService`
  * provides an easy way to resolve a SRV record, and then attempt to
  * connect to one of the hosts that implements that service, handling
  * service priority/weighting, multiple IP addresses, and multiple
  * address families.
  *
- * See #GSrvTarget for more information about SRV records, and see
- * #GSocketConnectable for an example of using the connectable
+ * See [struct@Gio.SrvTarget] for more information about SRV records, and see
+ * [iface@Gio.SocketConnectable] for an example of using the connectable
  * interface.
  */
 
-/**
- * GNetworkService:
- *
- * A #GSocketConnectable for resolving a SRV record and connecting to
- * that service.
- */
-
 struct _GNetworkServicePrivate
 {
   gchar *service, *protocol, *domain, *scheme;
@@ -123,34 +114,57 @@ g_network_service_class_init (GNetworkServiceClass *klass)
   gobject_class->get_property = g_network_service_get_property;
   gobject_class->finalize = g_network_service_finalize;
 
+  /**
+   * GNetworkService:service:
+   *
+   * Service name, for example `ldap`.
+   *
+   * Since: 2.22
+   */
   g_object_class_install_property (gobject_class, PROP_SERVICE,
-                                   g_param_spec_string ("service",
-                                                        P_("Service"),
-                                                        P_("Service name, eg \"ldap\""),
+                                   g_param_spec_string ("service", NULL, NULL,
                                                         NULL,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY |
                                                         G_PARAM_STATIC_STRINGS));
+
+  /**
+   * GNetworkService:protocol:
+   *
+   * Network protocol, for example `tcp`.
+   *
+   * Since: 2.22
+   */
   g_object_class_install_property (gobject_class, PROP_PROTOCOL,
-                                   g_param_spec_string ("protocol",
-                                                        P_("Protocol"),
-                                                        P_("Network protocol, eg \"tcp\""),
+                                   g_param_spec_string ("protocol", NULL, NULL,
                                                         NULL,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY |
                                                         G_PARAM_STATIC_STRINGS));
+
+  /**
+   * GNetworkService:domain:
+   *
+   * Network domain, for example `example.com`.
+   *
+   * Since: 2.22
+   */
   g_object_class_install_property (gobject_class, PROP_DOMAIN,
-                                   g_param_spec_string ("domain",
-                                                        P_("Domain"),
-                                                        P_("Network domain, eg, \"example.com\""),
+                                   g_param_spec_string ("domain", NULL, NULL,
                                                         NULL,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_CONSTRUCT_ONLY |
                                                         G_PARAM_STATIC_STRINGS));
+
+  /**
+   * GNetworkService:scheme:
+   *
+   * Network scheme (default is to use service).
+   *
+   * Since: 2.22
+   */
   g_object_class_install_property (gobject_class, PROP_DOMAIN,
-                                   g_param_spec_string ("scheme",
-                                                        P_("Scheme"),
-                                                        P_("Network scheme (default is to use service)"),
+                                   g_param_spec_string ("scheme", NULL, NULL,
                                                         NULL,
                                                         G_PARAM_READWRITE |
                                                         G_PARAM_STATIC_STRINGS));
@@ -364,17 +378,17 @@ static GList *
 g_network_service_fallback_targets (GNetworkService *srv)
 {
   GSrvTarget *target;
-  struct servent *entry;
+  gboolean has_port;
   guint16 port;
 
-  entry = getservbyname (srv->priv->service, "tcp");
-  port = entry ? g_ntohs (entry->s_port) : 0;
+  has_port = g_getservbyname_ntohs (srv->priv->service, "tcp", &port);
+
 #ifdef HAVE_ENDSERVENT
   endservent ();
 #endif
 
-  if (entry == NULL)
-      return NULL;
+  if (!has_port)
+    return NULL;
 
   target = g_srv_target_new (srv->priv->domain, port, 0, 0);
   return g_list_append (NULL, target);
